"""Define mixins to easily compose custom FilterDefinition classes."""


class TermsQueryMixin:
    """A mixin for filter definitions that need to apply term queries."""

    def get_query_fragment(self, data):
        """Build the query fragments as term queries for each selected value."""
        value_list = data.get(self.name)

        # For terms filters, as the name implies, it's a simple terms fragment
        return (
            [{"key": self.name, "fragment": [{"terms": {self.term: value_list}}]}]
            if value_list
            else []
        )


class ChoicesQueryMixin:
    """A mixin for filter definitions that need to apply predefined queries."""

    def get_query_fragment(self, data):
        """Pick the hardcoded query fragment for each selected value."""
        fragment_map = self.get_fragment_map()
        return [
            {"key": self.name, "fragment": fragment_map[value]}
            for value in data.get(self.name, [])
        ]


class ChoicesAggsMixin:
    """A mixin for filter definitions that need to apply aggregations for predefined choices."""

    # pylint: disable=unused-argument
    def get_aggs_fragment(self, queries, *args, **kwargs):
        """
        Build the aggregations as a set of filters, one for each possible value of the field.
        """
        return {
            # Create a custom aggregation for each possible choice for this filter
            # eg `availability@coming_soon` & `availability@current` & `availability@open`
            "{:s}@{:s}".format(self.name, choice_key): {
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries *but* the one(s) that
                        # filter on the current filter: we manually add back the only one that
                        # is relevant to the current choice.
                        "must": choice_fragment
                        + [
                            clause
                            for kf_pair in queries
                            for clause in kf_pair["fragment"]
                            if kf_pair["key"] is not self.name
                        ]
                    }
                }
            }
            for choice_key, choice_fragment in self.get_fragment_map().items()
        }


class NestedChoicesAggsMixin:
    """
    A mixin for filter definitions that are related to a nested field. The aggregation filter can
    only be recomputed at the level of the parent because it should group all queries of fields
    nested below the parent.
    """

    # pylint: disable=unused-argument
    def get_aggs_fragment(self, queries, data, parent, *args, **kwargs):
        """
        Computing aggregations for a nested field is DIFFICULT because query fragments related to
        nested fields are grouped under their common path. For example combined filters on
        availability and languages would lead to a query like:
            {
                "query": {
                    "nested": {
                        "path": "course_runs",
                        "query": {
                            "bool": {
                                "must": [
                                    {"range": {"course_runs.end": {"lte": "01-01-2019"}}},
                                    {"terms": {"course_runs.languages": ["de", "en", fr"]}},
                                ]
                            }
                        },
                    }
                }
            }
        In this example, computing the facet count for the French filter, is done with the
        following filter (excluding the filter on English and German so we only count French):
            {
                "query": {
                    "nested": {
                        "path": "course_runs",
                        "query": {
                            "bool": {
                                "must": [
                                    {"range": {"course_runs.end": {"lte": "01-01-2019"}}},
                                    {"terms": {"course_runs.languages": ["fr"]}},
                                ]
                            }
                        },
                    }
                }
            }

        This can only be built by calling the parent NestingWrapper with customized filter data.
        """
        return {
            # Create a custom aggregation for each possible choice for this filter
            # eg `availability@coming_soon` & `availability@current` & `availability@open`
            "{:s}@{:s}".format(self.name, choice_key): {
                "filter": {
                    "bool": {
                        # Use all the query fragments from the queries (the nesting parent is
                        # responsible for excluding the queries related to nested fields so we
                        # have to manually add them, making sure to apply on the current field
                        # only the current choice.
                        "must": [
                            clause
                            for kf_pair in (
                                queries
                                + parent.get_query_fragment(
                                    # override data with only the current choice
                                    {**data, self.name: [choice_key]}
                                )
                            )
                            for clause in kf_pair["fragment"]
                        ]
                    }
                }
            }
            for choice_key, choice_fragment in self.get_fragment_map().items()
        }



"""Django authentication backends.

For more information visit
https://docs.djangoproject.com/en/dev/topics/auth/customizing/.

This authentication backend is inspired by the official edX auth-backends
project:
https://github.com/edx/auth-backends/blob/master/auth_backends/backends.py
"""
import base64
import datetime
from calendar import timegm

from django.conf import settings
from django.core.cache import cache

from jose import jwk, jwt
from jose.jwt import ExpiredSignatureError, JWTClaimsError, JWTError
from social_core.backends.oauth import BaseOAuth2
from social_core.backends.open_id_connect import OpenIdConnectAuth
from social_core.exceptions import AuthTokenError

EDX_USER_PROFILE_TO_DJANGO = getattr(settings, "EDX_USER_PROFILE_TO_DJANGO", None) or {
    "preferred_username": "username",
    "email": "email",
    "name": "full_name",
    "given_name": "first_name",
    "family_name": "last_name",
    "locale": "language",
    "user_id": "user_id",
}


# pylint: disable=abstract-method,invalid-name
class EdXOIDC(OpenIdConnectAuth):
    """
    OIDC-based backend to authenticate with OpenEdx's LMS OpenID provider.

    DEPRECATED: this backend can be used with old OpenEdx releases, e.g.
    Dogwood.
    """

    name = "edx-oidc"

    DEFAULT_SCOPE = ["openid", "profile", "email"]
    ID_KEY = "preferred_username"
    REDIRECT_STATE = False

    def endpoint(self):
        """Get OIDC endpoint."""
        return self.setting("ENDPOINT").strip("/")

    @property
    def ID_TOKEN_ISSUER(self):
        """Build ID_TOKEN_ISSUER from the endpoint setting."""
        return self.endpoint()

    @property
    def ACCESS_TOKEN_URL(self):
        """Build ACCESS_TOKEN_URL from the endpoint setting."""
        return f"{self.endpoint()}/access_token/"

    @property
    def AUTHORIZATION_URL(self):
        """Build AUTHORIZATION_URL from the endpoint setting."""
        return f"{self.endpoint()}/authorize/"

    @property
    def USERINFO_URL(self):
        """Build USERINFO_URL from the endpoint setting."""
        return f"{self.endpoint()}/user_info/"

    def get_jwks_keys(self):
        """Returns the keys used to decode the ID token.

        Note: edX uses symmetric keys, so bypass the parent class's calls to an
        external server and return the key from settings.
        """
        k = self.setting("ID_TOKEN_DECRYPTION_KEY")
        k = base64.urlsafe_b64encode(k.encode("utf-8")).rstrip(b"=").decode("utf-8")
        return [{"k": k, "kty": "oct", "alg": "HS256"}]

    def validate_and_return_id_token(self, id_token, access_token):
        """
        Validates the id_token according to the steps at
        http://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation.
        """
        key = self.find_valid_key(id_token)

        if not key:
            raise AuthTokenError(self, "Signature verification failed")

        alg = key["alg"]
        rsa_key = jwk.construct(key)

        k = {
            "alg": rsa_key._algorithm,  # pylint: disable=protected-access
            "kty": "oct",
            "k": base64.urlsafe_b64encode(rsa_key.prepared_key)
            .rstrip(b"=")
            .decode("utf-8"),
        }

        try:
            claims = jwt.decode(
                id_token,
                k,
                algorithms=[alg],
                audience=self.setting("KEY"),
                issuer=self.id_token_issuer(),
                options=self.JWT_DECODE_OPTIONS,
            )
        except ExpiredSignatureError as error:
            raise AuthTokenError(self, "Signature has expired") from error
        except JWTClaimsError as error:
            raise AuthTokenError(self, str(error)) from error
        except JWTError as error:
            raise AuthTokenError(self, "Invalid signature") from error

        self.validate_claims(claims)

    def validate_claims(self, id_token):
        """Validate decoded JWT token."""

        if id_token["iss"] != self.id_token_issuer():
            raise AuthTokenError(self, "Invalid issuer")

        client_id = self.setting("KEY")

        if isinstance(id_token["aud"], str):
            id_token["aud"] = [id_token["aud"]]

        if client_id not in id_token["aud"]:
            raise AuthTokenError(self, "Invalid audience")

        if len(id_token["aud"]) > 1 and "azp" not in id_token:
            raise AuthTokenError(self, "Incorrect id_token: azp")

        if "azp" in id_token and id_token["azp"] != client_id:
            raise AuthTokenError(self, "Incorrect id_token: azp")

        utc_timestamp = timegm(datetime.datetime.utcnow().utctimetuple())
        if utc_timestamp > id_token["exp"]:
            raise AuthTokenError(self, "Signature has expired")

        if "nbf" in id_token and utc_timestamp < id_token["nbf"]:
            raise AuthTokenError(self, "Incorrect id_token: nbf")

        # Verify the token was issued in the last 10 minutes
        iat_leeway = self.setting("ID_TOKEN_MAX_AGE", self.ID_TOKEN_MAX_AGE)
        if utc_timestamp > id_token["iat"] + iat_leeway:
            raise AuthTokenError(self, "Incorrect id_token: iat")


# pylint: disable=abstract-method
class EdXOAuth2(BaseOAuth2):
    """
    OAuth2-based backend to authenticate with OpenEdx's LMS OAuth2 provider
    (hawthorn release).
    """

    name = "edx-oauth2"

    ACCESS_TOKEN_METHOD = "POST"  # nosec
    DEFAULT_SCOPE = ["profile", "email"]
    ID_KEY = "preferred_username"

    # EXTRA_DATA is used to store important data in the
    # UserSocialAuth.extra_data field. See:
    # https://python-social-auth.readthedocs.io/en/latest/backends/oauth.html?highlight=extra_data
    EXTRA_DATA = [
        # Update the stored user_id, if it's present in the response
        ("user_id", "user_id", True),
        # Update the stored refresh_token, if it's present in the response
        ("refresh_token", "refresh_token", True),
    ]

    def endpoint(self):
        """Get OAuth2 provider endpoint."""
        return self.setting("ENDPOINT").strip("/")

    @property
    def provider_configuration(self):
        """Get OAuth2 provider configuration and cache its configuration for a week."""

        cache_key = "edx_oauth2_provider_configuration"
        config = cache.get(cache_key)

        if not config:
            config = self.get_json(
                self.endpoint() + "/.well-known/openid-configuration"
            )

            # Cache for one week since the configuration rarely changes
            cache.set(
                cache_key,
                config,
                self.setting("PROVIDER_CONFIGURATION_CACHE_TTL", 604800),
            )

        return config

    def authorization_url(self):
        """Get OAuth2 provider authorization URL."""
        return self.provider_configuration.get("authorization_endpoint")

    def access_token_url(self):
        """Get OAuth2 provider access token URL."""
        return self.provider_configuration.get("token_endpoint")

    def end_session_url(self):
        """Get OAuth2 provider end session URL."""
        return self.provider_configuration.get("end_session_endpoint")

    def auth_complete_params(self, state=None):
        """Force the access token type to be JWT."""
        params = super().auth_complete_params(state)
        params.update({"token_type": "jwt"})
        return params

    def _get_jwks_keys(self):
        """Returns the keys used to decode the access token.

        Note: edX uses symmetric keys, so bypass the parent class's calls to an
        external server and return the key from settings.
        """
        k = self.setting("KEY")
        k = base64.urlsafe_b64encode(k.encode("utf-8")).rstrip(b"=").decode("utf-8")
        return [{"k": k, "kty": "oct", "alg": "HS256"}]

    def user_data(self, access_token, *args, **kwargs):
        """Get claimed user data from the JWT formatted access token."""
        decoded_access_token = jwt.decode(
            access_token,
            self._get_jwks_keys(),
            # We must skip verifications as edx does [1].
            # [1] https://github.com/edx/auth-backends/blob/6bf9d856c8e4cc4c1a72f67158468f8c94e3fca1/auth_backends/backends.py#L312 # noqa pylint: disable=line-too-long
            options={
                "verify_signature": False,
                "verify_aud": False,
                "verify_iat": False,
                "verify_exp": False,
                "verify_nbf": False,
                "verify_iss": False,
                "verify_sub": False,
                "verify_jti": False,
                "verify_at_hash": False,
                "leeway": 0,
            },
        )
        return {
            key: decoded_access_token[key]
            for key in EDX_USER_PROFILE_TO_DJANGO
            if key in decoded_access_token
        }

    def get_user_details(self, response):
        """Convert claim user details from the response."""
        return {
            d: response[s]
            for s, d in EDX_USER_PROFILE_TO_DJANGO.items()
            if s in response
        }



"""Declare and configure the models for richie's search application."""
from django.db import models


class SearchAccess(models.Model):
    """A model with no database table to hold global custom permissions."""

    class Meta:

        managed = False
        permissions = (
            ("can_manage_elasticsearch", "Allow managing Elasticsearch indices"),
        )



"""
Utility Django form fields that lets us validate every individual entry in an array,
using the Django-provided validator for that type of field
"""
from django import forms
from django.core.exceptions import ValidationError


class ArrayField(forms.Field):
    """
    Custom form field type that lets us validate and clean arrays of items, where both the array
    itself and each of its items are validated through the Django Form
    """

    def __init__(self, *args, **kwargs):
        """
        Get a hold of our custom arguments and delegate the rest to Field
        """
        # base_type is unknown to Field: pop it out of kwargs and keep it ourselves
        self.base_type = kwargs.pop("base_type")
        # Core arguments, including 'required', are left in the kwargs and handled by parent
        super().__init__(*args, **kwargs)

    def clean(self, value):
        """
        Validate our array and iterate over it, delegating item validation to the base_type that
        was passed to us by the consumer
        """
        if value:
            # We have something to check: attempt to iterate over it
            try:
                for base_type_value in value:
                    self.base_type.validate(base_type_value)
                return [
                    self.base_type.clean(base_type_value) for base_type_value in value
                ]
            # Iteration over subvalues failed, the value was invalid
            except TypeError as error:
                raise ValidationError(
                    "Failed to iterate over value, got a non-iterable type."
                ) from error
        # We're missing a required value
        elif self.required:
            raise ValidationError("Missing required value.")
        # Default to returning an empty array when an optional value is missing to simplify
        # the job for consumers of the form's data
        else:
            return []



"""
Settings specific to the courses application.
"""
from django.conf import settings
from django.utils.translation import gettext_lazy as _

ADMIN = "ADMIN"

ROLE_CHOICES = ((ADMIN, _("Admin")),)

COURSE_ADMIN_ROLE = {
    "django_permissions": [
        # Objects
        "cms.add_page",
        "cms.change_page",
        "cms.add_title",
        "cms.change_title",
        "courses.add_courserun",
        "courses.change_courserun",
        "courses.add_person",
        "courses.change_person",
        "cms.use_structure",
        # Filer
        "filer.add_file",
        "filer.change_file",
        "filer.view_file",
        "filer.add_image",
        "filer.change_image",
        "filer.view_image",
        # Plugins
        "djangocms_link.add_link",
        "djangocms_link.change_link",
        "djangocms_link.delete_link",
        "djangocms_link.view_link",
        "djangocms_picture.add_picture",
        "djangocms_picture.change_picture",
        "djangocms_picture.delete_picture",
        "djangocms_picture.view_picture",
        "djangocms_text_ckeditor.add_text",
        "djangocms_text_ckeditor.change_text",
        "djangocms_text_ckeditor.delete_text",
        "djangocms_text_ckeditor.view_text",
        "djangocms_video.add_videoplayer",
        "djangocms_video.change_videoplayer",
        "djangocms_video.delete_videoplayer",
        "djangocms_video.view_videoplayer",
        "djangocms_video.add_videosource",
        "djangocms_video.change_videosource",
        "djangocms_video.delete_videosource",
        "djangocms_video.view_videosource",
        "djangocms_video.add_videotrack",
        "djangocms_video.change_videotrack",
        "djangocms_video.delete_videotrack",
        "djangocms_video.view_videotrack",
        "plain_text.add_plaintext",
        "plain_text.change_plaintext",
        "plain_text.delete_plaintext",
        "plain_text.view_plaintext",
        "section.add_section",
        "section.change_section",
        "section.delete_section",
        "section.view_section",
        "simple_text_ckeditor.add_simpletext",
        "simple_text_ckeditor.change_simpletext",
        "simple_text_ckeditor.delete_simpletext",
        "simple_text_ckeditor.view_simpletext",
        "courses.add_organizationpluginmodel",
        "courses.change_organizationpluginmodel",
        "courses.delete_organizationpluginmodel",
        "courses.view_organizationpluginmodel",
        "courses.add_personpluginmodel",
        "courses.change_personpluginmodel",
        "courses.delete_personpluginmodel",
        "courses.view_personpluginmodel",
        "courses.add_categorypluginmodel",
        "courses.change_categorypluginmodel",
        "courses.delete_categorypluginmodel",
        "courses.view_categorypluginmodel",
        "courses.add_licencepluginmodel",
        "courses.change_licencepluginmodel",
        "courses.delete_licencepluginmodel",
        "courses.view_licencepluginmodel",
    ],
    "course_page_permissions": {
        "can_change": True,
        "can_add": True,
        "can_delete": False,
        "can_change_advanced_settings": False,
        "can_publish": False,
        "can_change_permissions": False,
        "can_move_page": True,
        "can_view": False,  # beware: can_view = True makes it a view restriction...
        "grant_on": 5,  # page and descendants
    },
    "course_folder_permissions": {
        "can_read": True,
        "can_edit": False,
        "can_add_children": True,
        "type": 2,  # folder and children
    },
}
COURSE_ADMIN_ROLE.update(getattr(settings, "RICHIE_COURSE_ADMIN_ROLE", {}))

ORGANIZATION_ADMIN_ROLE = {
    "django_permissions": [
        # Objects
        "cms.add_page",
        "cms.change_page",
        "cms.add_title",
        "cms.change_title",
        "courses.add_course",
        "courses.change_course",
        "courses.add_courserun",
        "courses.change_courserun",
        "courses.add_person",
        "courses.change_person",
        "cms.use_structure",
        # Filer
        "filer.add_file",
        "filer.change_file",
        "filer.view_file",
        "filer.add_image",
        "filer.change_image",
        "filer.view_image",
        # Plugins
        "djangocms_link.add_link",
        "djangocms_link.change_link",
        "djangocms_link.delete_link",
        "djangocms_link.view_link",
        "djangocms_picture.add_picture",
        "djangocms_picture.change_picture",
        "djangocms_picture.delete_picture",
        "djangocms_picture.view_picture",
        "djangocms_text_ckeditor.add_text",
        "djangocms_text_ckeditor.change_text",
        "djangocms_text_ckeditor.delete_text",
        "djangocms_text_ckeditor.view_text",
        "djangocms_video.add_videoplayer",
        "djangocms_video.change_videoplayer",
        "djangocms_video.delete_videoplayer",
        "djangocms_video.view_videoplayer",
        "djangocms_video.add_videosource",
        "djangocms_video.change_videosource",
        "djangocms_video.delete_videosource",
        "djangocms_video.view_videosource",
        "djangocms_video.add_videotrack",
        "djangocms_video.change_videotrack",
        "djangocms_video.delete_videotrack",
        "djangocms_video.view_videotrack",
        "plain_text.add_plaintext",
        "plain_text.change_plaintext",
        "plain_text.delete_plaintext",
        "plain_text.view_plaintext",
        "section.add_section",
        "section.change_section",
        "section.delete_section",
        "section.view_section",
        "simple_text_ckeditor.add_simpletext",
        "simple_text_ckeditor.change_simpletext",
        "simple_text_ckeditor.delete_simpletext",
        "simple_text_ckeditor.view_simpletext",
        "courses.add_organizationpluginmodel",
        "courses.change_organizationpluginmodel",
        "courses.delete_organizationpluginmodel",
        "courses.view_organizationpluginmodel",
        "courses.add_personpluginmodel",
        "courses.change_personpluginmodel",
        "courses.delete_personpluginmodel",
        "courses.view_personpluginmodel",
        "courses.add_categorypluginmodel",
        "courses.change_categorypluginmodel",
        "courses.delete_categorypluginmodel",
        "courses.view_categorypluginmodel",
        "courses.add_licencepluginmodel",
        "courses.change_licencepluginmodel",
        "courses.delete_licencepluginmodel",
        "courses.view_licencepluginmodel",
    ],
    "organization_page_permissions": {
        "can_change": True,
        "can_add": True,
        "can_delete": False,
        "can_change_advanced_settings": False,
        "can_publish": False,
        "can_change_permissions": False,
        "can_move_page": False,
        "can_view": False,  # beware: can_view = True makes it a view restriction...
        "grant_on": 1,  # just the page
    },
    "organization_folder_permissions": {
        "can_read": True,
        "can_edit": False,
        "can_add_children": True,
        "type": 2,  # folder and children
    },
    "courses_page_permissions": {
        "can_change": True,
        "can_add": True,
        "can_delete": False,
        "can_change_advanced_settings": False,
        "can_publish": False,
        "can_change_permissions": False,
        "can_move_page": True,
        "can_view": False,  # beware: can_view = True makes it a view restriction...
        "grant_on": 5,  # page and descendants
    },
    "courses_folder_permissions": {
        "can_read": True,
        "can_edit": False,
        "can_add_children": True,
        "type": 2,  # folder and children
    },
}
ORGANIZATION_ADMIN_ROLE.update(getattr(settings, "RICHIE_ORGANIZATION_ADMIN_ROLE", {}))

PAGE_EXTENSION_TOOLBAR_ITEM_POSITION = getattr(
    settings, "RICHIE_PAGE_EXTENSION_TOOLBAR_ITEM_POSITION", 4
)

# Glimpses

BLOGPOST_GLIMPSE_VARIANT_CHOICES = getattr(
    settings,
    "RICHIE_BLOGPOST_GLIMPSE_VARIANT_CHOICES",
    [
        (None, _("Inherit")),
        ("glimpse", _("Default")),
        ("mini", _("Mini")),
        ("favorite", _("Favorite")),
    ],
)
CATEGORY_GLIMPSE_VARIANT_CHOICES = getattr(
    settings,
    "RICHIE_CATEGORY_GLIMPSE_VARIANT_CHOICES",
    [
        (None, _("Inherit")),
        ("glimpse", _("Default")),
        ("badge", _("badge")),
        ("tag", _("Tag")),
    ],
)
COURSE_GLIMPSE_VARIANT_CHOICES = getattr(
    settings,
    "RICHIE_COURSE_GLIMPSE_VARIANT_CHOICES",
    [(None, _("Inherit")), ("glimpse", _("Default")), ("small", "small")],
)

ORGANIZATION_GLIMPSE_VARIANT_CHOICES = getattr(
    settings,
    "RICHIE_ORGANIZATION_GLIMPSE_VARIANT_CHOICES",
    [
        (None, _("Inherit")),
        ("glimpse", _("Default")),
        ("card", _("Card")),
        ("row", _("Row")),
    ],
)

# For each type of page we define:
#   - the `reverse_id` of the page under which pages should be created via the wizard,
#   - the template to be used when creating a new page of this type,
BLOGPOSTS_PAGE = {
    "reverse_id": "blogposts",
    "template": "courses/cms/blogpost_detail.html",
}
CATEGORIES_PAGE = {
    "reverse_id": "categories",
    "template": "courses/cms/category_detail.html",
}
COURSERUNS_PAGE = {"template": "courses/cms/course_run_detail.html"}
COURSES_PAGE = {"reverse_id": "courses", "template": "courses/cms/course_detail.html"}
ORGANIZATIONS_PAGE = {
    "reverse_id": "organizations",
    "template": "courses/cms/organization_detail.html",
}
PERSONS_PAGE = {"reverse_id": "persons", "template": "courses/cms/person_detail.html"}
PROGRAMS_PAGE = {
    "reverse_id": "programs",
    "template": "courses/cms/program_detail.html",
}

PAGES_INFO = {
    "home": {
        "title": "Home",
        "in_navigation": False,
        "is_homepage": True,
        "template": "richie/homepage.html",
    },
    BLOGPOSTS_PAGE["reverse_id"]: {
        "title": "News",
        "in_navigation": True,
        "template": "courses/cms/blogpost_list.html",
    },
    CATEGORIES_PAGE["reverse_id"]: {
        "title": "Categories",
        "in_navigation": True,
        "template": "courses/cms/category_list.html",
    },
    COURSES_PAGE["reverse_id"]: {
        "title": "Courses",
        "in_navigation": True,
        "template": "search/search.html",
    },
    ORGANIZATIONS_PAGE["reverse_id"]: {
        "title": "Organizations",
        "in_navigation": True,
        "template": "courses/cms/organization_list.html",
    },
    PERSONS_PAGE["reverse_id"]: {
        "title": "Persons",
        "in_navigation": True,
        "template": "courses/cms/person_list.html",
    },
    PROGRAMS_PAGE["reverse_id"]: {
        "title": "Programs",
        "in_navigation": True,
        "template": "courses/cms/program_list.html",
    },
}
PAGES_INFO.update(getattr(settings, "PAGES_INFO", {}))

ROOT_REVERSE_IDS = PAGES_INFO.keys()

# Fields effort and duration
MINUTE, HOUR, DAY, WEEK, MONTH = "minute", "hour", "day", "week", "month"
DEFAULT_TIME_UNIT = HOUR
DEFAULT_EFFORT_UNIT = WEEK
DEFAULT_REFERENCE_UNIT = MONTH

TIME_UNITS = {
    MINUTE: (_("minute"), _("minutes")),
    HOUR: (_("hour"), _("hours")),
    DAY: (_("day"), _("days")),
    WEEK: (_("week"), _("weeks")),
    MONTH: (_("month"), _("months")),
}



"""
API routes exposed by our Courses app.
"""
from rest_framework import routers

from .api import CourseRunsViewSet

# Instantiate a router
ROUTER = routers.SimpleRouter()

# Define our app's routes with the router
ROUTER.register(r"course-runs", CourseRunsViewSet, "course-runs")

# Use the standard name for our urlpatterns so urls.py can import it effortlessly
urlpatterns = ROUTER.urls



"""
Section CMS plugin
"""
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from richie.apps.core.defaults import PLUGINS_GROUP

from .defaults import SECTION_TEMPLATES
from .forms import SectionForm
from .models import Section


@plugin_pool.register_plugin
class SectionPlugin(CMSPluginBase):
    """
    CMSPlugin to add a content section with a distinct title from content.
    """

    allow_children = True
    cache = True
    form = SectionForm
    fieldsets = ((None, {"fields": ["title", "template"]}),)
    model = Section
    module = PLUGINS_GROUP
    name = _("Section")

    # Required from CMSPluginBase signature but not used since we override it
    # from render()
    render_template = SECTION_TEMPLATES[0][0]

    def render(self, context, instance, placeholder):
        context.update({"instance": instance, "placeholder": placeholder})
        self.render_template = instance.template
        return context



"""SimplePicture plugin forms."""
from django import forms

from djangocms_picture.models import Picture


class SimplePictureForm(forms.ModelForm):
    """Plugin form used to add a picture from the frontend admin."""

    class Meta:
        """Form meta attributes"""

        model = Picture
        fields = ["picture"]



"""
PlainText CMS plugin factories
"""
import factory
from djangocms_picture.models import Picture

from richie.apps.core.factories import FilerImageFactory


class PictureFactory(factory.django.DjangoModelFactory):
    """
    Factory to create random instances of SimplePicture for testing.
    """

    class Meta:
        model = Picture

    picture = factory.SubFactory(FilerImageFactory)



"""
NestedItem CMS plugin factories
"""
import factory

from .defaults import NESTEDITEM_VARIANTS
from .models import NestedItem


class NestedItemFactory(factory.django.DjangoModelFactory):
    """
    Factory to create random instances of NestedItem for testing.
    """

    class Meta:
        model = NestedItem

    variant = NESTEDITEM_VARIANTS[0][0]
    content = factory.Faker("text", max_nb_chars=84)



"""
API endpoints for the courses app.
"""
from rest_framework.permissions import BasePermission
from rest_framework.viewsets import ModelViewSet

from .models import CourseRun
from .serializers import CourseRunSerializer


class NotAllowed(BasePermission):
    """
    Utility permission class to deny all requests. This is used as a default to close
    requests to unsupported actions.
    """

    def has_permission(self, request, view):
        """
        Always deny permission.
        """
        return False


class CourseRunsViewSet(ModelViewSet):
    """
    API endpoints to access and perform actions on course runs.
    """

    permission_classes = [NotAllowed]
    queryset = CourseRun.objects.all()
    serializer_class = CourseRunSerializer

    def get_permissions(self):
        """
        Manage permissions for builtin DRF methods on ViewSets.
        """
        if self.action == "retrieve":
            permission_classes = []
        else:
            try:
                permission_classes = getattr(self, self.action).kwargs.get(
                    "permission_classes"
                )
            except AttributeError:
                permission_classes = self.permission_classes

        return [permission() for permission in permission_classes]



# Generated by Django 2.1.12 on 2019-09-06 10:02

import django.db.models.deletion
from django.db import migrations, models

import richie.apps.core.models


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("courses", "0005_auto_20190717_0827"),
    ]

    operations = [
        migrations.CreateModel(
            name="Program",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.Program",
                    ),
                ),
            ],
            options={
                "verbose_name": "program",
                "db_table": "richie_program",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="ProgramPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_programpluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "program__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="program_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "program plugin",
                "db_table": "richie_program_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
    ]



"""Update Elasticsearch indices each time a page is modified."""
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from richie.apps.courses.models import Category, Course, Organization
from richie.apps.search.index_manager import richie_bulk
from richie.apps.search.indexers import ES_INDICES


def update_course(instance, _language):
    """
    Update Elasticsearch indices when a course was modified and published:
    - update the course document in the Elasticsearch courses index.

    Returns None if the page was related to a course and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a course.
    """
    course = Course.objects.get(draft_extension__extended_object=instance)
    richie_bulk([ES_INDICES.courses.get_es_document_for_course(course)])


def update_course_run(instance, _language):
    """
    Update Elasticsearch indices when a course run was modified and published:
    - update the course document in the Elasticsearch courses index for the parent course
      of this course run.

    Returns None if the page was related to a course run and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a course run.
    """
    course = instance.courserun.get_course().public_extension
    richie_bulk([ES_INDICES.courses.get_es_document_for_course(course)])


def update_organization(instance, language):
    """
    Update Elasticsearch indices when an organization was modified and published:
    - update the organization document in the Elasticsearch organizations index for the
      organization and its direct parent (because the parent ID may change from Parent to Leaf),
    - update the course documents in the Elasticsearch courses index for all courses linked to
      this organization.

    Returns None if the page was related to an organization and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to an organization.
    """
    organization = Organization.objects.get(draft_extension__extended_object=instance)
    actions = [
        ES_INDICES.courses.get_es_document_for_course(course)
        for course in organization.get_courses(language)
    ]
    actions.append(
        ES_INDICES.organizations.get_es_document_for_organization(organization)
    )

    # Update the organization's parent only if it exists
    try:
        parent = organization.extended_object.get_parent_page().organization
    except AttributeError:
        pass
    else:
        actions.append(
            ES_INDICES.organizations.get_es_document_for_organization(parent)
        )

    richie_bulk(actions)


def update_category(instance, language):
    """
    Update Elasticsearch indices when a category was modified and published:
    - update the category document in the Elasticsearch categories index for the category
      and its direct parent (because the parent ID may change from Parent to Leaf),
    - update the course documents in the Elasticsearch courses index for all courses linked to
      this category.

    Returns None if the page was related to a category and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a category.
    """
    category = Category.objects.get(draft_extension__extended_object=instance)
    actions = [
        ES_INDICES.courses.get_es_document_for_course(course)
        for course in category.get_courses(language)
    ]
    actions.append(ES_INDICES.categories.get_es_document_for_category(category))

    # Update the category's parent only if it exists
    try:
        parent = category.extended_object.get_parent_page().category
    except AttributeError:
        pass
    else:
        actions.append(ES_INDICES.categories.get_es_document_for_category(parent))

    richie_bulk(actions)


def update_page_extension(instance, language):
    """
    Try updating each type of page extension one-by-one until one works (because we don't know
    to which type of page extension this page is related).
    """
    for method in [
        update_course,
        update_course_run,
        update_category,
        update_organization,
    ]:
        try:
            # The method should raise an ObjectDoesNotExist exception if the page extension
            # linked to this instance is of another type.
            method(instance, language)
        except ObjectDoesNotExist:
            continue
        else:
            return


# pylint: disable=unused-argument
def on_page_publish(sender, instance, language, **kwargs):
    """
    Trigger update of the Elasticsearch indices impacted by the modification of the instance
    only once the database transaction is successful.
    """
    if getattr(settings, "RICHIE_KEEP_SEARCH_UPDATED", True):
        transaction.on_commit(lambda: update_page_extension(instance, language))



"""Default settings for the core app of Richie."""
from django.conf import global_settings, settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import lazy
from django.utils.translation import gettext_lazy as _

# Group to add plugin to placeholder "Content"
PLUGINS_GROUP = _("Richie Plugins")

# Django sets `LANGUAGES` by default with all supported languages. We can use it for example for
# the choice of languages on the course run which should not be limited to the few languages
# active in the CMS.
# pylint: disable=no-member
ALL_LANGUAGES = getattr(
    settings,
    "ALL_LANGUAGES",
    [(language, _(name)) for language, name in global_settings.LANGUAGES],
)
ALL_LANGUAGES_DICT = dict(ALL_LANGUAGES)

# Careful! Languages should be ordered by priority, as this tuple is used to get
# fallback/default languages throughout the app.
# Use "en" as default as it is the language that is most likely to be spoken by any visitor
# when their preferred language, whatever it is, is unavailable
LANGUAGES_DICT = dict(settings.LANGUAGES)

# check that all languages in LANGUAGES are matching one of the languages in ALL_LANGUAGES
for site_language in LANGUAGES_DICT:
    generic_language_code = site_language.split("-")[0]
    for language in ALL_LANGUAGES_DICT:
        if language.startswith(generic_language_code):
            break
    else:
        raise ImproperlyConfigured(
            f'"{site_language:s}" in LANGUAGES does not match any language in ALL_LANGUAGES'
        )

# Our frontend expects BCP47/RFC5646 locales (eg. fr-FR). Django, on the other hand,
# supports ISO639-1 language codes without a region (e.g. fr) which is sufficient
# if you don't need to differentiate several regional versions of the same language.
#
# BCP47/RFC5646 lets us specify language-region pairs by combining ISO639-1 language codes and
# ISO3166-1 region codes using a dash (eg. fr-FR).
#
# The only thing we're missing for this to be bridged completely is to make sure locales
# set on the <html> element are also available in our React frontend.
#
# This is where this setting comes in: it lists available locales by order of priority:
# the first matching locale will be used if your LANGUAGES setting declares languages without
# regions.
RFC_5646_LOCALES = lazy(
    lambda: getattr(settings, "RFC_5646_LOCALES", ["en-US", "es-ES", "fr-FR", "fr-CA"]),
    list,
)()

GLIMPSE_PAGINATION = {
    "categories": 8,
    "courses": 8,
    "organizations": 8,
    "blogposts": 6,
    "persons": 4,
    "programs": 4,
}



"""
Backend to connect Open edX richie with an LMS
"""
import logging

import requests
from requests.auth import AuthBase

from .base import BaseLMSBackend

logger = logging.getLogger(__name__)


class EdXTokenAuth(AuthBase):
    """Attach HTTP token authentication to the given Request object."""

    def __init__(self, token):
        """Set-up token value in the instance."""
        self.token = token

    def __call__(self, request):
        """Modify and return the request."""
        request.headers.update(
            {"X-Edx-Api-Key": self.token, "Content-Type": "application/json"}
        )
        return request


class TokenAPIClient(requests.Session):
    """
    A :class:`requests.Session` that automatically authenticates against edX's preferred
    authentication method up to Dogwood, given a secret token.

    For more usage details, see documentation of the :class:`requests.Session` object:
    https://requests.readthedocs.io/en/master/user/advanced/#session-objects
    """

    def __init__(self, token, *args, **kwargs):
        """Extending the session object by setting the authentication token."""
        super().__init__(*args, **kwargs)
        self.auth = EdXTokenAuth(token)


class TokenEdXLMSBackend(BaseLMSBackend):
    """LMS backend for Richie tested with Open EdX Dogwood to Hawthorn."""

    @property
    def api_client(self):
        """Instantiate and return an edx token API client."""
        return TokenAPIClient(self.configuration["API_TOKEN"])



# Generated by Django 2.2.12 on 2020-04-17 10:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nesteditem", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="nesteditem",
            name="content",
            field=models.TextField(blank=True, default="", verbose_name="Content"),
        ),
    ]



"""
Base backend to connect richie with an LMS
"""


class BaseLMSBackend:
    """
    Base backend to hold the methods common to all backends and provide a skeleton for others.
    """

    def __init__(self, configuration, *args, **kwargs):
        """Attache configuration to the backend instance."""
        super().__init__(*args, **kwargs)
        self.configuration = configuration



"""
Simple text plugin models
"""
from django.db import models
from django.utils.encoding import force_str
from django.utils.html import strip_tags
from django.utils.text import Truncator
from django.utils.translation import gettext_lazy as _

from cms.models.pluginmodel import CMSPlugin
from djangocms_text_ckeditor.html import clean_html


class SimpleText(CMSPlugin):
    """
    Simple text plugin model.

    Aim to be used instead of TextPlugin which produces orphan plugin item that
    broke some behavior with constraints from ``CMS_PLACEHOLDER_CONF``.
    """

    body = models.TextField(_("body"))

    def __str__(self):
        return Truncator(strip_tags(self.body)).words(6, truncate="...")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.body = force_str(self.body)

    # pylint: disable=signature-differs
    def save(self, *args, **kwargs):
        # Clean HTML from potential XSS content
        self.body = clean_html(self.body, full=False)

        super().save(*args, **kwargs)



# Generated by Django 2.2.7 on 2019-11-19 23:14
from django.db import migrations

from cms.api import add_plugin
from cms.models import StaticPlaceholder


def migrate_footer_to_static_placeholder(apps, schema_editor):
    """
    Create a footer with the new static placeholder from the existing footer pages that were
    placed under the "annex" page and displayed in the footer via a `show_menu_below_id`
    template tag.
    """
    Page = apps.get_model("cms", "Page")
    Title = apps.get_model("cms", "Title")

    # We should import StaticPlaceholder from apps but its `draft` and `public` fields
    # are custom foreign key field that checks that they are targeting an instance of
    # `cms.models.Placeholder` so the code would not work. We can safely assume that the
    # Placeholder and StaticPlaceholder models are still there when this migration is run
    static_placeholder, was_created = StaticPlaceholder.objects.get_or_create(
        code="footer"
    )

    if not was_created:
        # If the static placeholder was already existing, it means this migration is being
        # replayed and we better do nothing.
        return

    for is_draft in [False, True]:
        # Look for an existing footer page
        try:
            footer_page = Page.objects.get(
                reverse_id="annex", publisher_is_draft=is_draft
            )
        except Page.DoesNotExist:
            return

        placeholder = (
            static_placeholder.draft if is_draft else static_placeholder.public
        )

        for language in Title.objects.filter(page=footer_page).values_list(
            "language", flat=True
        ):
            # Create the <ul> section to carry the list of links
            section = add_plugin(
                placeholder,
                plugin_type="SectionPlugin",
                language=language,
                template="richie/section/section_list.html",
            )

            # Create a <li> link for each page in the exiting footer menu
            for page in Page.objects.filter(
                node__parent=footer_page.node,
                in_navigation=True,
                title_set__language=language,
                publisher_is_draft=is_draft,
            ):
                title = page.title_set.get(language=language)
                add_plugin(
                    placeholder,
                    plugin_type="LinkPlugin",
                    language=language,
                    internal_link_id=page.id,
                    name=title.title,
                    target=section,
                )


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0009_auto_20191014_1801"),
        ("section", "0003_auto_20191119_1650"),
    ]

    operations = [
        migrations.RunPython(
            migrate_footer_to_static_placeholder, migrations.RunPython.noop
        )
    ]



"""Create_demo_site management command."""
import logging
import random
from collections import defaultdict

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand, CommandError
from django.test.utils import override_settings

import factory
from cms.api import add_plugin
from cms.models import StaticPlaceholder

from richie.apps.core.factories import create_text_plugin, image_getter
from richie.apps.core.helpers import recursive_page_creation
from richie.apps.courses import factories
from richie.plugins.glimpse import defaults as glimpse_defaults
from richie.plugins.glimpse.factories import GlimpseFactory

from ... import defaults
from ...helpers import create_categories
from ...utils import pick_image

logger = logging.getLogger("richie.commands.demo.create_demo_site")


def get_number_of_course_runs():
    """
    Returns a random integer between 1 and the max number of course runs.
    We make it a convenience method so that it can be mocked in tests.
    """
    return random.randint(1, defaults.NB_OBJECTS["course_courseruns"])  # nosec


def get_number_of_icons():
    """
    Returns a random integer between 0 and the max number of course icons.
    0 is weighted to have as much chance to be chosen as all other numbers cumulated because
    we visually don't want all our courses to have an icon.
    We make it a convenience method so that it can be mocked in tests.
    """
    return random.choice(  # nosec
        [0] * defaults.NB_OBJECTS["course_icons"]
        + list(range(defaults.NB_OBJECTS["course_icons"]))
    )


# pylint: disable=no-member
#
# Looks like pylint is not relevant at guessing object types when cascading
# methods over querysets: Instance of 'list' has no 'delete' member (no-member).
# We choose to ignore this false positive warning.

# pylint: disable=too-many-locals,too-many-statements,too-many-branches


@override_settings(RICHIE_KEEP_SEARCH_UPDATED=False)
def create_demo_site():
    """
    Create a simple site tree structure for developpers to work in realistic environment.

    We create multilingual pages, add organizations under the related page and add
    plugins to each page.
    """
    site = Site.objects.get(id=1)
    site.domain = getattr(
        settings, "RICHIE_DEMO_SITE_DOMAIN", defaults.DEFAULT_DEMO_SITE_DOMAIN
    )
    site.name = "Richie demonstration"
    site.save()

    lms_endpoint = getattr(
        settings, "LMS_BACKENDS", [{"BASE_URL": defaults.DEFAULT_LMS_ENDPOINT}]
    )[0]["BASE_URL"]

    # Create pages as described in PAGES_INFOS
    pages_created = recursive_page_creation(site, defaults.PAGES_INFO)

    # Create the footer links
    def create_footer_link(**link_info):
        """
        Use LinkPlugin to create a link in footer menu with link_info

        Links can be nested into a NestedItemPlugin, in this case link_info contains
        a target key.
        """
        if "internal_link" in link_info:
            link_info = link_info.copy()
            link_info["internal_link"] = pages_created[link_info["internal_link"]]
        add_plugin(
            plugin_type="LinkPlugin",
            **link_info,
        )

    footer_static_ph = StaticPlaceholder.objects.get_or_create(code="footer")[0]
    for footer_placeholder in [footer_static_ph.draft, footer_static_ph.public]:
        for language, content in defaults.FOOTER_CONTENT.items():
            for footer_info in content:
                if "items" in footer_info:
                    # Create the first level items for main columns
                    nest_column_plugin = add_plugin(
                        footer_placeholder,
                        plugin_type="NestedItemPlugin",
                        language=language,
                        content=footer_info.get("title", ""),
                    )

                    # Create the second level items for links
                    for item_info in footer_info.get("items", []):
                        create_footer_link(
                            language=language,
                            placeholder=footer_placeholder,
                            target=nest_column_plugin,
                            **item_info,
                        )
                else:
                    # Create link at first level
                    create_footer_link(
                        language=language,
                        placeholder=footer_placeholder,
                        **footer_info,
                    )

    # Create some licences
    licences = (
        factories.LicenceFactory.create_batch(
            defaults.NB_OBJECTS["licences"],
            logo__file__from_path=pick_image("licence")(),
        )
        if defaults.NB_OBJECTS.get("licences")
        else []
    )

    # Generate each category tree and return a list of the leaf categories
    icons = list(
        create_categories(
            **defaults.ICONS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("category_logo"),
            page_parent=pages_created["categories"],
        )
    )
    levels = list(
        create_categories(
            **defaults.LEVELS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("category_logo"),
            page_parent=pages_created["categories"],
        )
    )
    subjects = list(
        create_categories(
            **defaults.SUBJECTS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("category_logo"),
            page_parent=pages_created["categories"],
        )
    )
    partnerships = list(
        create_categories(
            **defaults.PARTNERSHIPS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("category_logo"),
            page_parent=pages_created["categories"],
        )
    )
    tags = list(
        create_categories(
            **defaults.TAGS_INFO,
            fill_banner=pick_image("banner"),
            fill_logo=pick_image("category_logo"),
            page_parent=pages_created["categories"],
        )
    )

    # Create organizations under the `Organizations` page
    organizations = []
    for i in range(defaults.NB_OBJECTS["organizations"]):
        # Randomly assign each organization to a partnership level category
        organizations.append(
            factories.OrganizationFactory(
                page_in_navigation=True,
                page_languages=["en", "fr"],
                page_parent=pages_created["organizations"],
                fill_banner=pick_image("banner"),
                fill_categories=[random.choice(partnerships)]  # nosec
                if (i % 2 == 0)
                else [],
                fill_description=True,
                fill_logo=pick_image("logo"),
                should_publish=True,
                with_permissions=True,
            )
        )

    # Create persons under the `persons` page
    persons = []
    persons_for_organization = defaultdict(list)
    for _ in range(defaults.NB_OBJECTS["persons"]):
        # Randomly assign each person to a set of organizations
        person_organizations = random.sample(
            organizations,
            random.randint(1, defaults.NB_OBJECTS["person_organizations"]),  # nosec
        )
        person = factories.PersonFactory(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["persons"],
            fill_categories=random.sample(
                subjects,
                random.randint(1, defaults.NB_OBJECTS["person_subjects"]),  # nosec
            ),
            fill_organizations=person_organizations,
            fill_portrait=pick_image("portrait"),
            fill_bio=True,
            fill_maincontent=True,
            should_publish=True,
        )
        persons.append(person)
        for organization in person_organizations:
            persons_for_organization[organization.id].append(person)

    # Assign each person randomly to an organization so that our course are tagged realistically
    # If organizations and persons are tagged randomly on courses, each organizations will
    # in the end be related to most persons... not what we want.

    # Create courses under the `Course` page with categories and organizations
    # relations
    courses = []
    for _ in range(defaults.NB_OBJECTS["courses"]):
        video_sample = random.choice(factories.VIDEO_SAMPLE_LINKS)  # nosec

        # Randomly assign each course to a set of organizations
        course_organizations = random.sample(
            organizations, defaults.NB_OBJECTS["course_organizations"]
        )

        # Only the persons members of these organizations are eligible to be part
        # of the course team
        eligible_persons = set(
            person
            for o in course_organizations
            for person in persons_for_organization[o.id]
        )

        course_licences = (
            [
                ("course_license_content", random.choice(licences)),  # nosec
                ("course_license_participation", random.choice(licences)),  # nosec
            ]
            if licences
            else []
        )

        course = factories.CourseFactory(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["courses"],
            fill_licences=course_licences,
            fill_team=random.sample(
                eligible_persons,
                min(
                    random.randint(1, defaults.NB_OBJECTS["course_persons"]),  # nosec
                    len(eligible_persons),
                ),
            ),
            fill_teaser=video_sample,
            fill_cover=pick_image("cover")(video_sample.image),
            fill_categories=[
                *random.sample(
                    subjects,
                    random.randint(1, defaults.NB_OBJECTS["course_subjects"]),  # nosec
                ),
                random.choice(levels),  # nosec
            ],
            fill_icons=random.sample(icons, get_number_of_icons()),
            fill_organizations=course_organizations,
            fill_plan=True,
            fill_texts=[
                "course_assessment",
                "course_description",
                "course_introduction",
                "course_format",
                "course_prerequisites",
                "course_skills",
            ],
        )
        course.create_permissions_for_organization(course_organizations[0])
        courses.append(course)

        # Add extra information
        for language in course.extended_object.get_languages():
            placeholder = course.extended_object.placeholders.get(
                slot="course_information"
            )
            nb_half_rows = 2 * random.randint(0, 1)  # nosec
            nb_full_rows = random.randint(0, 2)  # nosec
            nb_cards = 4 * random.randint(0, 1)  # nosec

            # Partners
            if nb_half_rows or nb_full_rows:
                partner_section = add_plugin(
                    language=language,
                    placeholder=placeholder,
                    plugin_type="SectionPlugin",
                    title=defaults.COURSE_CONTENT[language]["partners_title"],
                )
            for _ in range(nb_half_rows):
                glimpse_data = factory.build(
                    dict,
                    FACTORY_CLASS=GlimpseFactory,
                    variant=glimpse_defaults.ROW_HALF,
                    image=image_getter(pick_image("logo")()),
                )
                glimpse_data["image"].save()
                add_plugin(
                    language=language,
                    placeholder=placeholder,
                    plugin_type="GlimpsePlugin",
                    target=partner_section,
                    **glimpse_data,
                )
            for _ in range(nb_full_rows):
                glimpse_data = factory.build(
                    dict,
                    FACTORY_CLASS=GlimpseFactory,
                    variant=glimpse_defaults.ROW_FULL,
                    image=image_getter(pick_image("logo")()),
                )
                glimpse_data["image"].save()
                add_plugin(
                    language=language,
                    placeholder=placeholder,
                    plugin_type="GlimpsePlugin",
                    target=partner_section,
                    **glimpse_data,
                )
            # Sponsors
            if nb_cards:
                sponsor_section = add_plugin(
                    language=language,
                    placeholder=placeholder,
                    plugin_type="SectionPlugin",
                    title=defaults.COURSE_CONTENT[language]["sponsors_title"],
                )
            for _ in range(nb_cards):
                glimpse_data = factory.build(
                    dict,
                    FACTORY_CLASS=GlimpseFactory,
                    variant=glimpse_defaults.CARD_SQUARE,
                    image=image_getter(pick_image("logo")()),
                )
                glimpse_data["image"].save()
                add_plugin(
                    language=language,
                    placeholder=placeholder,
                    plugin_type="GlimpsePlugin",
                    target=sponsor_section,
                    **glimpse_data,
                )

        for language in course.extended_object.get_languages():
            course.extended_object.publish(language)

        # Add a random number of course runs to the course
        nb_course_runs = get_number_of_course_runs()
        # pick a subset of languages for this course (otherwise all courses will have more or
        # less all the languages across their course runs!)
        languages_subset = random.sample(
            ["de", "en", "es", "fr", "it", "nl"], random.randint(1, 4)  # nosec
        )
        for i in range(nb_course_runs):
            factories.CourseRunFactory(
                __sequence=i,
                languages=random.sample(
                    languages_subset, random.randint(1, len(languages_subset))  # nosec
                ),
                page_in_navigation=False,
                page_languages=["en", "fr"],
                page_parent=course.extended_object,
                resource_link=f"{lms_endpoint}/courses/course-v1:edX+DemoX+Demo_Course/info",
                should_publish=True,
            )

    # Create blog posts under the `News` page
    blogposts = []
    for _ in range(defaults.NB_OBJECTS["blogposts"]):
        post = factories.BlogPostFactory.create(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["blogposts"],
            fill_cover=pick_image("cover"),
            fill_excerpt=True,
            fill_body=True,
            fill_categories=[
                *random.sample(levels, defaults.NB_OBJECTS["blogpost_levels"]),
                *random.sample(tags, defaults.NB_OBJECTS["blogpost_tags"]),
            ],
            fill_author=random.sample(persons, 1),
            should_publish=True,
        )
        blogposts.append(post)

    # Create programs under the `Programs` page
    programs = []
    for _ in range(defaults.NB_OBJECTS["programs"]):
        program = factories.ProgramFactory.create(
            page_in_navigation=True,
            page_languages=["en", "fr"],
            page_parent=pages_created["programs"],
            fill_cover=pick_image("cover"),
            fill_excerpt=True,
            fill_body=True,
            fill_courses=[
                *random.sample(courses, defaults.NB_OBJECTS["programs_courses"])
            ],
            should_publish=True,
        )
        programs.append(program)

    # Create some content on the programs list page
    placeholder = pages_created["programs"].placeholders.get(slot="maincontent")

    for language in pages_created["programs"].get_languages():
        create_text_plugin(
            pages_created["programs"],
            placeholder,
            nb_paragraphs=random.randint(3, 4),  # nosec
            languages=[language],
            plugin_type="TextPlugin",
        )

        # Once content has been added we must publish again homepage
        pages_created["programs"].publish(language)

    # Once everything has been created, use some content to create a homepage
    placeholder = pages_created["home"].placeholders.get(slot="maincontent")

    # - Get a banner image
    banner = image_getter(pick_image("banner")())

    # - Get a logo image
    logo = image_getter(pick_image("logo")())

    # - Create the home page in each language
    for language, content in defaults.HOMEPAGE_CONTENT.items():
        # Add a banner
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LargeBannerPlugin",
            title=content["banner_title"],
            background_image=banner,
            content=content["banner_content"],
            template=content["banner_template"],
        )
        # Add highlighted courses with a button
        courses_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["courses_title"],
            template=content["section_template"],
        )
        for course in random.sample(courses, defaults.NB_OBJECTS["home_courses"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CoursePlugin",
                target=courses_section,
                page=course.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=courses_section,
            name=content["courses_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["courses"],
        )

        # Add highlighted blogposts
        blogposts_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["blogposts_title"],
            template=content["section_template"],
        )
        for blogpost in random.sample(blogposts, defaults.NB_OBJECTS["home_blogposts"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="BlogPostPlugin",
                target=blogposts_section,
                page=blogpost.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=blogposts_section,
            name=content["blogposts_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["blogposts"],
        )

        # Add highlighted programs
        programs_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["programs_title"],
            template=content["section_template"],
        )
        for program in random.sample(programs, defaults.NB_OBJECTS["home_programs"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="ProgramPlugin",
                target=programs_section,
                page=program.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=programs_section,
            name=content["programs_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["programs"],
        )

        # Add highlighted organizations
        organizations_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["organizations_title"],
            template=content["section_template"],
        )
        for organization in random.sample(
            organizations, defaults.NB_OBJECTS["home_organizations"]
        ):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="OrganizationPlugin",
                target=organizations_section,
                page=organization.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=organizations_section,
            name=content["organizations_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["organizations"],
        )

        # Add highlighted subjects
        subjects_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["subjects_title"],
            template=content["section_template"],
        )
        for subject in random.sample(subjects, defaults.NB_OBJECTS["home_subjects"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="CategoryPlugin",
                target=subjects_section,
                page=subject.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=subjects_section,
            name=content["subjects_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["categories"],
        )

        # Add highlighted persons
        persons_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["persons_title"],
            template=content["section_template"],
        )
        for person in random.sample(persons, defaults.NB_OBJECTS["home_persons"]):
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="PersonPlugin",
                target=persons_section,
                page=person.extended_object,
            )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=persons_section,
            name=content["persons_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["persons"],
        )

        # Add Glimpse quotes with empty title
        quotes_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title="",
        )
        for _ in range(3):
            glimpse_data = factory.build(
                dict,
                FACTORY_CLASS=GlimpseFactory,
                variant=glimpse_defaults.QUOTE,
                title=None,
                image=image_getter(pick_image("portrait")()),
            )
            glimpse_data["image"].save()
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="GlimpsePlugin",
                target=quotes_section,
                **glimpse_data,
            )

        # Add Glimpse cards with empty content
        cards_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title="",
        )
        for _ in range(4):
            glimpse_data = factory.build(
                dict,
                FACTORY_CLASS=GlimpseFactory,
                variant=glimpse_defaults.CARD_SQUARE,
                content=None,
                image=image_getter(pick_image("cover")()),
            )
            glimpse_data["image"].save()
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="GlimpsePlugin",
                target=cards_section,
                **glimpse_data,
            )

        # Once content has been added we must publish again homepage
        pages_created["home"].publish(language)

    # Fill the single column sample page
    placeholder = pages_created["annex__about"].placeholders.get(slot="maincontent")

    # - Get a banner image
    banner = image_getter(pick_image("banner")())

    # - Get a logo image
    logo = image_getter(pick_image("logo")())

    # - Get a video
    video_sample = random.choice(factories.VIDEO_SAMPLE_LINKS)  # nosec

    # - Create sample page in each language
    for language, content in defaults.SINGLECOLUMN_CONTENT.items():
        # Add a banner
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LargeBannerPlugin",
            title=content["banner_title"],
            background_image=banner,
            content=content["banner_content"],
            template=content["banner_template"],
        )
        # HTML paragraphs
        create_text_plugin(
            pages_created["annex__about"],
            placeholder,
            nb_paragraphs=random.randint(3, 4),  # nosec
            languages=[language],
            plugin_type="TextPlugin",
        )
        # A large video sample
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="VideoPlayerPlugin",
            label=video_sample.label,
            embed_link=video_sample.url,
            template="full-width",
        )
        # Section with some various plugins
        sample_section = add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SectionPlugin",
            title=content["section_sample_title"],
            template=content["section_sample_template"],
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="OrganizationPlugin",
            target=sample_section,
            page=random.choice(organizations).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="CoursePlugin",
            target=sample_section,
            page=random.choice(courses).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="OrganizationPlugin",
            target=sample_section,
            page=random.choice(organizations).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="BlogPostPlugin",
            target=sample_section,
            page=random.choice(blogposts).extended_object,  # nosec
        )
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="LinkPlugin",
            target=sample_section,
            name=content["section_sample_button_title"],
            template=content["button_template_name"],
            internal_link=pages_created["home"],
        )
        # Add a licence
        if licences:
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="LicencePlugin",
                licence=random.choice(licences),  # nosec
            )
        # Add a simple picture entry
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="SimplePicturePlugin",
            picture=logo,
        )
        # Add a plain text
        text = factory.Faker(
            "text", max_nb_chars=random.randint(150, 250)  # nosec
        ).generate({"locale": language})
        add_plugin(
            language=language,
            placeholder=placeholder,
            plugin_type="PlainTextPlugin",
            body=text,
        )

        # Once content has been added we must publish again the about page
        pages_created["annex__about"].publish(language)

    # Create a sitemap page
    placeholder = pages_created["annex__sitemap"].placeholders.get(slot="maincontent")

    for language in pages_created["annex__sitemap"].get_languages():
        parent_instance = add_plugin(
            language=language, placeholder=placeholder, plugin_type="HTMLSitemapPlugin"
        )
        for name, params in defaults.SITEMAP_PAGE_PARAMS.items():
            add_plugin(
                language=language,
                placeholder=placeholder,
                plugin_type="HTMLSitemapPagePlugin",
                target=parent_instance,
                root_page=pages_created[name],
                **params,
            )

        # Once content has been added we must publish again the sitemap
        pages_created["annex__sitemap"].publish(language)


class Command(BaseCommand):
    """Create default pages for FUN frontend"""

    help = __doc__

    def add_arguments(self, parser):

        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="Force command execution despite DEBUG is set to False",
        )

    def handle(self, *args, **options):

        if not settings.DEBUG and not options["force"]:
            raise CommandError(
                (
                    "This command is not meant to be used in production environment "
                    "except you know what you are doing, if so use --force parameter"
                )
            )

        create_demo_site()

        logger.info("done")



"""
Simple text plugin forms
"""
from html import unescape
from unicodedata import normalize

from django import forms
from django.utils.html import strip_spaces_between_tags

from djangocms_text_ckeditor.widgets import TextEditorWidget

from .models import SimpleText


class CKEditorPluginForm(forms.ModelForm):
    """
    Plugin form used to fill its content from frontend admin.
    """

    class Meta:
        """
        Form meta attributes
        """

        model = SimpleText
        widgets = {"body": TextEditorWidget}
        fields = ["body"]

    def clean_body(self):
        """Normalize and unescape the text submitted by CKEditor then remove useless spaces."""
        body = self.cleaned_data.get("body", "")
        body = normalize("NFKC", body)
        body = unescape(body)
        return strip_spaces_between_tags(body)



# Generated by Django 2.1.7 on 2019-03-26 09:08

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import filer.fields.image
import parler.models

import richie.apps.core.fields.multiselect
import richie.apps.core.models
import richie.apps.courses.models.category


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        migrations.swappable_dependency(settings.FILER_IMAGE_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BlogPost",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.BlogPost",
                    ),
                ),
            ],
            options={
                "verbose_name": "blog post",
                "db_table": "richie_blog_post",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="BlogPostPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_blogpostpluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "blogpost__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="blogpost_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "blog post plugin",
                "db_table": "richie_blog_post_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.Category",
                    ),
                ),
            ],
            options={
                "verbose_name": "category",
                "db_table": "richie_category",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="CategoryPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_categorypluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to=richie.apps.courses.models.category.get_category_limit_choices_to,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="category_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "category plugin",
                "db_table": "richie_category_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="Course",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.Course",
                    ),
                ),
            ],
            options={"verbose_name": "course", "db_table": "richie_course"},
        ),
        migrations.CreateModel(
            name="CoursePluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_coursepluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "course__isnull": False,
                            "node__parent__cms_pages__course__isnull": True,
                            "node__parent__cms_pages__publisher_is_draft": True,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="course_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "course plugin",
                "db_table": "richie_course_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="CourseRun",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "resource_link",
                    models.URLField(
                        blank=True, null=True, verbose_name="Resource link"
                    ),
                ),
                (
                    "start",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="course start"
                    ),
                ),
                (
                    "end",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="course end"
                    ),
                ),
                (
                    "enrollment_start",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="enrollment start"
                    ),
                ),
                (
                    "enrollment_end",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="enrollment end"
                    ),
                ),
                (
                    "languages",
                    richie.apps.core.fields.multiselect.MultiSelectField(
                        choices=[
                            ("af", "Afrikaans"),
                            ("ar", "Arabic"),
                            ("ast", "Asturian"),
                            ("az", "Azerbaijani"),
                            ("bg", "Bulgarian"),
                            ("be", "Belarusian"),
                            ("bn", "Bengali"),
                            ("br", "Breton"),
                            ("bs", "Bosnian"),
                            ("ca", "Catalan"),
                            ("cs", "Czech"),
                            ("cy", "Welsh"),
                            ("da", "Danish"),
                            ("de", "German"),
                            ("dsb", "Lower Sorbian"),
                            ("el", "Greek"),
                            ("en", "English"),
                            ("en-au", "Australian English"),
                            ("en-gb", "British English"),
                            ("eo", "Esperanto"),
                            ("es", "Spanish"),
                            ("es-ar", "Argentinian Spanish"),
                            ("es-co", "Colombian Spanish"),
                            ("es-mx", "Mexican Spanish"),
                            ("es-ni", "Nicaraguan Spanish"),
                            ("es-ve", "Venezuelan Spanish"),
                            ("et", "Estonian"),
                            ("eu", "Basque"),
                            ("fa", "Persian"),
                            ("fi", "Finnish"),
                            ("fr", "French"),
                            ("fy", "Frisian"),
                            ("ga", "Irish"),
                            ("gd", "Scottish Gaelic"),
                            ("gl", "Galician"),
                            ("he", "Hebrew"),
                            ("hi", "Hindi"),
                            ("hr", "Croatian"),
                            ("hsb", "Upper Sorbian"),
                            ("hu", "Hungarian"),
                            ("ia", "Interlingua"),
                            ("id", "Indonesian"),
                            ("io", "Ido"),
                            ("is", "Icelandic"),
                            ("it", "Italian"),
                            ("ja", "Japanese"),
                            ("ka", "Georgian"),
                            ("kab", "Kabyle"),
                            ("kk", "Kazakh"),
                            ("km", "Khmer"),
                            ("kn", "Kannada"),
                            ("ko", "Korean"),
                            ("lb", "Luxembourgish"),
                            ("lt", "Lithuanian"),
                            ("lv", "Latvian"),
                            ("mk", "Macedonian"),
                            ("ml", "Malayalam"),
                            ("mn", "Mongolian"),
                            ("mr", "Marathi"),
                            ("my", "Burmese"),
                            ("nb", "Norwegian Bokmål"),
                            ("ne", "Nepali"),
                            ("nl", "Dutch"),
                            ("nn", "Norwegian Nynorsk"),
                            ("os", "Ossetic"),
                            ("pa", "Punjabi"),
                            ("pl", "Polish"),
                            ("pt", "Portuguese"),
                            ("pt-br", "Brazilian Portuguese"),
                            ("ro", "Romanian"),
                            ("ru", "Russian"),
                            ("sk", "Slovak"),
                            ("sl", "Slovenian"),
                            ("sq", "Albanian"),
                            ("sr", "Serbian"),
                            ("sr-latn", "Serbian Latin"),
                            ("sv", "Swedish"),
                            ("sw", "Swahili"),
                            ("ta", "Tamil"),
                            ("te", "Telugu"),
                            ("th", "Thai"),
                            ("tr", "Turkish"),
                            ("tt", "Tatar"),
                            ("udm", "Udmurt"),
                            ("uk", "Ukrainian"),
                            ("ur", "Urdu"),
                            ("vi", "Vietnamese"),
                            ("zh-hans", "Simplified Chinese"),
                            ("zh-hant", "Traditional Chinese"),
                        ],
                        help_text="The list of languages in which the course content is available.",
                        max_choices=50,
                        max_length=255,
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.CourseRun",
                    ),
                ),
            ],
            options={"verbose_name": "course run", "db_table": "richie_course_run"},
        ),
        migrations.CreateModel(
            name="Licence",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200, verbose_name="name")),
                (
                    "url",
                    models.CharField(blank=True, max_length=255, verbose_name="url"),
                ),
                ("content", models.TextField(default="", verbose_name="content")),
                (
                    "logo",
                    filer.fields.image.FilerImageField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="licence",
                        to=settings.FILER_IMAGE_MODEL,
                        verbose_name="logo",
                    ),
                ),
            ],
            options={"verbose_name": "licence", "db_table": "richie_licence"},
        ),
        migrations.CreateModel(
            name="LicencePluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_licencepluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, default="", verbose_name="description"
                    ),
                ),
                (
                    "licence",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="courses.Licence",
                    ),
                ),
            ],
            options={
                "verbose_name": "licence plugin",
                "db_table": "richie_licence_plugin",
            },
            bases=("cms.cmsplugin",),
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        max_length=100,
                        null=True,
                        verbose_name="code",
                    ),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
                (
                    "public_extension",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="draft_extension",
                        to="courses.Organization",
                    ),
                ),
            ],
            options={
                "verbose_name": "organization",
                "db_table": "richie_organization",
                "ordering": ["-pk"],
            },
        ),
        migrations.CreateModel(
            name="OrganizationPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_organizationpluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "organization__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="organization_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "organization plugin",
                "db_table": "richie_organization_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(max_length=200, verbose_name="First name"),
                ),
                (
                    "last_name",
                    models.CharField(max_length=200, verbose_name="Last name"),
                ),
                (
                    "extended_object",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
            ],
            options={"verbose_name": "person", "db_table": "richie_person"},
        ),
        migrations.CreateModel(
            name="PersonPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_personpluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "person__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "person plugin",
                "db_table": "richie_person_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.CreateModel(
            name="PersonTitle",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                )
            ],
            options={"verbose_name": "person title", "db_table": "richie_person_title"},
            bases=(parler.models.TranslatableModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="PersonTitleTranslation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language_code",
                    models.CharField(
                        db_index=True, max_length=15, verbose_name="Language"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="Title")),
                (
                    "abbreviation",
                    models.CharField(max_length=10, verbose_name="Title abbreviation"),
                ),
                (
                    "master",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="translations",
                        to="courses.PersonTitle",
                    ),
                ),
            ],
            options={
                "verbose_name": "person title translation",
                "db_table": "richie_person_title_translation",
            },
        ),
        migrations.AddField(
            model_name="person",
            name="person_title",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="persons",
                to="courses.PersonTitle",
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="public_extension",
            field=models.OneToOneField(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="draft_extension",
                to="courses.Person",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="persontitletranslation", unique_together={("language_code", "master")}
        ),
    ]



"""Utility functions for admin classes."""
from django.urls import reverse
from django.utils.html import format_html


def link_field(field_name, view_name=None, anchor=None, target_blank=True):
    """Convert the representation of an object linked by a foreign key into a clickable link.

    Parameters
    ----------
    field_name (string):
        If `field_name` is "name", link text will be str(obj.name) and link will be the admin
        url for obj.name.id:change.
    view_name (string):
        Optional alternative view name to compute the url of the linked object in place of the
        standard object admin change form.
    anchor (string):
        Optional alternative anchor to display on the link in place of the str representation
        of the linked object.

    Returns
    -------
    function
        The function that Django admin must call with the object as argument to render the field
        as a link.

    """

    def _link_field(obj):
        """Render a link in Django admin for foreign key fields.

        The link replaces the string representation of the linked object that is rendered
        by Django by default for foreign keys.

        Parameters
        ----------
        obj: Type[models.Model]
            The instance of Django model for which we want to render the field `field_name`.

        Returns
        -------
        string
            The html representing the link to the object admin change view.

        """
        linked_obj = getattr(obj, field_name)
        if linked_obj is None:
            return "-"

        local_view_name = view_name
        if not local_view_name:
            app_label = linked_obj._meta.app_label
            model_name = linked_obj._meta.model_name
            local_view_name = f"admin:{app_label}_{model_name}_change"

        local_anchor = anchor or str(linked_obj)
        link_url = reverse(local_view_name, args=[linked_obj.id])
        markup = (
            '<a target="_blank" href="{:s}">{!s}</a>'
            if target_blank
            else '<a href="{:s}">{!s}</a>'
        )
        return format_html(markup, link_url, local_anchor)

    _link_field.short_description = field_name
    return _link_field



"""Common helpers for different kinds of filter definitions."""
from ..defaults import FACET_COUNTS_DEFAULT_LIMIT, FACET_COUNTS_MAX_LIMIT


def applicable_facet_limit(data, filter_name):
    """
    Detect the applicable facet counts limit for a given filter depending on the request params.
    It should be `FACET_COUNTS_DEFAULT_LIMIT` — which is intended to keep a clean UI on the search
    engine — unless the request is for a specific subset of values, in which case we increase the
    limit to `FACET_COUNTS_MAX_LIMIT` — which exists to provide a sensible upper boundary.
    """
    try:
        has_self_include = data[f"{filter_name:s}_include"]
        return (
            FACET_COUNTS_MAX_LIMIT if has_self_include else FACET_COUNTS_DEFAULT_LIMIT
        )
    except KeyError:
        return FACET_COUNTS_DEFAULT_LIMIT



"""Custom validators for the simple text editor plugin."""
from html import unescape
from unicodedata import normalize

from django.core.validators import BaseValidator
from django.utils.deconstruct import deconstructible
from django.utils.html import strip_spaces_between_tags, strip_tags
from django.utils.translation import ngettext_lazy


@deconstructible
class HTMLMaxLengthValidator(BaseValidator):
    """
    A custom validator that limits the max length of an HTML content, counting only the real
    content, and not the HTML tags and useless spaces.
    """

    message = ngettext_lazy(
        "Ensure this text has at most %(limit_value)d character (it has %(show_value)d).",
        "Ensure this text has at most %(limit_value)d characters (it has %(show_value)d).",
        "limit_value",
    )
    code = "max_length_html"

    def compare(self, a, b):
        """The validator should raise the error if the text length is more than the limit."""
        return a > b

    def clean(self, x):
        """
        Before counting the number of characters in the text submitted by CKEditor:
        - normalize it,
        - unescape it,
        - strip all HTML tags,
        - strip useless spaces.
        """
        x = normalize("NFKC", unescape(x))
        x = strip_tags(strip_spaces_between_tags(x))
        return len(x)



"""SimplePicture plugin for DjangoCMS."""
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from djangocms_picture.models import Picture

from richie.apps.core.defaults import PLUGINS_GROUP

from .forms import SimplePictureForm
from .helpers import get_picture_info


@plugin_pool.register_plugin
class SimplePicturePlugin(CMSPluginBase):
    """
    CMSPlugin to easily add an image when all display options for the image are determined by
    code i.e via presets defined in the project settings or via rendering contexts in templates.
    """

    allow_children = False
    cache = True
    disable_child_plugins = True
    form = SimplePictureForm
    model = Picture
    module = PLUGINS_GROUP
    name = _("Image")
    render_template = "richie/simple_picture/picture.html"
    render_plugin = True

    fieldsets = ((None, {"fields": ["picture"]}),)

    def render(self, context, instance, placeholder):
        """
        Compute thumbnails and populate the context with all the information necessary to display
        the image as defined in settings.

        The image is chosen in the following order:
          - preset matching a name passed in the context via the `picture_preset` parameter,
          - preset matching the name of the placeholder in which the plugin is being used,
          - fallback to the `default` preset.
        """
        # Look for the name of a preset in the context and default to the name of the placeholder
        preset_name = context.get("picture_preset", placeholder)

        context["picture_info"] = get_picture_info(instance, preset_name)
        context["instance"] = instance
        return context



"""SimplePicture plugin for DjangoCMS."""
from .defaults import SIMPLEPICTURE_PRESETS


def get_picture_info(instance, preset_name):
    """
    Compute picture information for a given preset defined in settings.

    A preset is of the form:

    "default": {
        "src": {"size": (1000, 1000), "crop": "smart"}, # easythumbnail options
        "srcset": [
            {
                "options": {"size": (800, 800), "crop": "smart"},
                "descriptor": "800w",
            },
            {
                "options": {"size": (600, 600), "crop": "smart"},
                "descriptor": "600w",
            },
        ],
        "sizes": "100vw", # e.g 1000px or 100vw
    }

    """
    # Bail out if the picture does not have an image as that's the object we use to get
    # all the information we need to return any picture info.
    if not instance.picture:
        return None

    thumbnailer = instance.picture.easy_thumbnails_thumbnailer

    # Look for the preset in settings and fallback to "default"
    preset = SIMPLEPICTURE_PRESETS.get(preset_name, SIMPLEPICTURE_PRESETS["default"])

    # Complete picture information with thumbnails url calculated according to what is
    # defined in the preset
    picture_info = {}
    location_dict = {"subject_location": instance.picture.subject_location}

    # - src
    options = preset["src"].copy()
    options.update(location_dict)
    picture_info["src"] = thumbnailer.get_thumbnail(options).url

    # - srcset
    srcset = []
    for info in preset.get("srcset", []):
        options = info["options"].copy()
        options.update(location_dict)
        srcset.append(
            "{:s} {:s}".format(
                thumbnailer.get_thumbnail(options).url, info["descriptor"]
            )
        )
    picture_info["srcset"] = ", ".join(srcset) if srcset else None

    # - sizes
    picture_info["sizes"] = preset.get("sizes")

    return picture_info



"""Serializers for Richie's courses app."""
from rest_framework import serializers

from .models import CourseRun


class CourseRunSerializer(serializers.ModelSerializer):
    """
    Course run serializer. Includes state but not any nested object or CMS page information.
    """

    class Meta:
        model = CourseRun
        fields = [
            "id",
            "resource_link",
            "start",
            "end",
            "enrollment_start",
            "enrollment_end",
            "languages",
            "state",
        ]



# Generated by Django 2.1.12 on 2019-09-30 20:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("courses", "0006_add_program")]

    operations = [
        migrations.AddField(
            model_name="organizationpluginmodel",
            name="variant",
            field=models.CharField(
                blank=True,
                choices=[(None, "Default")],
                help_text="Optional glimpse form factor for custom look.",
                max_length=50,
                null=True,
                verbose_name="variant",
            ),
        )
    ]



# Generated by Django 2.1.12 on 2019-10-01 10:12

import django.db.models.deletion
from django.db import migrations, models

import richie.apps.core.models


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("courses", "0007_auto_20190930_2245"),
    ]

    operations = [
        migrations.CreateModel(
            name="OrganizationsByCategoryPluginModel",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="courses_organizationsbycategorypluginmodel",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "variant",
                    models.CharField(
                        blank=True,
                        choices=[(None, "Default")],
                        help_text="Optional glimpse variant for a custom look.",
                        max_length=50,
                        null=True,
                        verbose_name="variant",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        limit_choices_to={
                            "category__isnull": False,
                            "publisher_is_draft": True,
                        },
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="organizations_by_category_plugins",
                        to="cms.Page",
                    ),
                ),
            ],
            options={
                "verbose_name": "organizations by category plugin",
                "db_table": "richie_organizations_by_category_plugin",
            },
            bases=(richie.apps.core.models.PagePluginMixin, "cms.cmsplugin"),
        ),
        migrations.AlterField(
            model_name="organizationpluginmodel",
            name="variant",
            field=models.CharField(
                blank=True,
                choices=[(None, "Default")],
                help_text="Optional glimpse variant for a custom look.",
                max_length=50,
                null=True,
                verbose_name="variant",
            ),
        ),
    ]



"""
Large banner plugin models
"""
from django.db import models
from django.utils.text import Truncator
from django.utils.translation import gettext_lazy as _

from cms.models.pluginmodel import CMSPlugin
from filer.fields.image import FilerImageField

from .defaults import LARGEBANNER_TEMPLATES


# pylint: disable=model-no-explicit-unicode,line-too-long
#
# We choose to ignore the 'model-no-explicit-unicode' since we are running
# python 3, and pylint-django fails at detecting the python 2 compatibility
# layer [1] handled by Django CMS [2]. And we also ignore the 'line-too-long'
# warning caused by reference urls.
#
# [1] https://docs.djangoproject.com/en/1.11/ref/utils/#django.utils.encoding.python_2_unicode_compatible # noqa
# [2] https://github.com/divio/django-cms/blob/3.5.2/cms/models/pluginmodel.py#L171
class LargeBanner(CMSPlugin):
    """
    Model to configure a home page banner with background image, logo and title.
    """

    title = models.CharField(max_length=255)
    background_image = FilerImageField(
        related_name="background_image",
        verbose_name=_("background image"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    logo = FilerImageField(
        related_name="logo",
        verbose_name=_("logo"),
        on_delete=models.PROTECT,
        null=True,
        blank=True,
    )
    logo_alt_text = models.CharField(max_length=255, blank=True)
    template = models.CharField(
        _("Template"),
        max_length=150,
        choices=LARGEBANNER_TEMPLATES,
        default=LARGEBANNER_TEMPLATES[0][0],
        blank=False,
        help_text=_("Choose template to render plugin."),
    )
    content = models.TextField(_("Content"), blank=True, default="")

    def __str__(self):
        return Truncator(self.title).words(6, truncate="...")



"""
Template context processors
"""
import json

from django.conf import settings
from django.contrib.sites.models import Site
from django.middleware.csrf import get_token

from . import defaults


def site_metas(request):
    """
    Context processor to add all information required by Richie CMS templates and frontend.

    If `CDN_DOMAIN` settings is defined we add it in the context. It allows
    to load statics js on a CDN like cloudfront.
    """
    site_current = Site.objects.get_current()
    protocol = "https" if request.is_secure() else "http"
    authentication_delegation = getattr(settings, "AUTHENTICATION_DELEGATION", {})
    context = {
        **{
            f"GLIMPSE_PAGINATION_{k.upper()}": v
            for k, v in {
                **defaults.GLIMPSE_PAGINATION,
                **getattr(settings, "RICHIE_GLIMPSE_PAGINATION", {}),
            }.items()
        },
        "SITE": {
            "name": site_current.name,
            "domain": site_current.domain,
            "web_url": f"{protocol:s}://{site_current.domain:s}",
        },
        "AUTHENTICATION": {
            "PROFILE_URLS": json.dumps(
                [
                    {
                        "label": str(url["label"]),
                        "action": str(
                            url["href"].format(
                                base_url=authentication_delegation["BASE_URL"]
                            )
                        ),
                    }
                    for url in authentication_delegation["PROFILE_URLS"]
                ]
            ),
        },
        "FRONTEND_CONTEXT": json.dumps(
            {
                "context": {
                    "csrftoken": get_token(request),
                    "environment": getattr(settings, "ENVIRONMENT", ""),
                    "release": getattr(settings, "RELEASE", ""),
                    "sentry_dsn": getattr(settings, "SENTRY_DSN", ""),
                    "authentication": {
                        "endpoint": authentication_delegation["BASE_URL"],
                        "backend": authentication_delegation["BACKEND"],
                    },
                    "lms_backends": [
                        {
                            "endpoint": lms["BASE_URL"],
                            "backend": lms["BACKEND"],
                            "course_regexp": lms["JS_COURSE_REGEX"],
                            "selector_regexp": lms["JS_SELECTOR_REGEX"],
                        }
                        for lms in getattr(settings, "LMS_BACKENDS", [])
                    ],
                }
            }
        ),
    }
    if getattr(settings, "CDN_DOMAIN", False):
        context["CDN_DOMAIN"] = settings.CDN_DOMAIN

    return context



"""HTML Sitemap plugin for DjangoCMS."""
from functools import reduce

from django.contrib.sites.models import Site
from django.db.models import F, Q
from django.utils import translation
from django.utils.translation import gettext_lazy as _

from cms.constants import PUBLISHER_STATE_PENDING
from cms.models.pagemodel import Page
from cms.models.pluginmodel import CMSPlugin
from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from .models import HTMLSitemapPage


def annotate_pages(pages):
    """A generator to annotate pages with nesting information.

    This is required to build a nested list of pages from the flat mptree list of pages.

    Returns:
        tuple(page, nb_open, nb_close, is_leaf):
            - page: the current page to wrap in a <li> tag,
            - nb_open(int): number of opening <ul> tags. Should be 1 when the nesting level
                increases, 0 otherwise,
            - nb_close(int): number of closing </ul> tags. Should be the gap between one page
                and the next when less nested. When reaching the last page we should close all
                the tags still open.
            - is_leaf(boolean): True if the current <li> tag is a leaf, False otherwise. If
                the page is not empty, it has children and we should delay closing the <li>.
    """
    nb_open = nb_close = 0
    nb_pages = len(pages)

    for i in range(nb_pages):
        if i == 0:
            nb_open = 0
        else:
            nb_open = max(pages[i].level - pages[i - 1].level, 0)

        is_leaf = True
        if i == nb_pages - 1:
            nb_close = pages[i].level - 1
        else:
            gap = pages[i].level - pages[i + 1].level
            nb_close = max(gap, 0)
            is_leaf = gap >= 0

        yield pages[i], nb_open, nb_close, is_leaf


@plugin_pool.register_plugin
class HTMLSitemapPlugin(CMSPluginBase):
    """HTML sitemap parent plugin acting as a container for sitemap pages."""

    model = CMSPlugin
    name = _("HTML Sitemap")
    render_template = "richie/html_sitemap/html_sitemap.html"
    allow_children = True
    child_classes = ["HTMLSitemapPagePlugin"]
    cache = False

    @classmethod
    def get_empty_change_form_text(cls, obj=None):
        """
        Customize the message on the empty change form to explain that children plugins must
        be added.
        """
        return _(
            "Press save to create a site map. You will then be able to add a child plugin for "
            "each subtree in your sitemap."
        )


@plugin_pool.register_plugin
class HTMLSitemapPagePlugin(CMSPluginBase):
    """HTML sitemap pages to add a subtree to a parent sitemap plugin."""

    model = HTMLSitemapPage
    name = _("HTML sitemap page")
    render_template = "richie/html_sitemap/html_sitemap_item.html"
    require_parent = True
    parent_classes = ["HTMLSitemapPlugin"]
    cache = True

    def render(self, context, instance, placeholder):
        """
        Prepare the list of pages to display in the sitemap and add it to the context.

        The list of pages if annotated with information about nesting so we are able
        to construct the <ul> <li> structure in the template.
        """
        language = translation.get_language()
        current_page_is_draft = instance.placeholder.page.publisher_is_draft

        if instance.root_page:
            if current_page_is_draft:
                root_page = instance.root_page
            else:
                root_page = instance.root_page.get_public_object()

            # Get all descendants of the root page including the root page itself
            node = root_page.node
            pages = Page.objects.filter(node__in=node.__class__.get_tree(node))
            depth = node.depth if instance.include_root_page else node.depth + 1

            if not instance.include_root_page:
                pages = pages.exclude(id=root_page.id)

        else:
            # Get all pages of the current site from its root
            pages = Page.objects.filter(node__site=Site.objects.get_current())
            depth = 1

        if instance.max_depth:
            pages = pages.filter(node__depth__lte=depth + instance.max_depth - 1)

        pages = (
            pages.filter(publisher_is_draft=current_page_is_draft, is_page_type=False)
            .select_related("node")
            .order_by("node__path")
        )

        exclusion_queries = []

        # For anonymous users, exclude pages that require login and their descendants
        if not context["request"].user.is_authenticated:
            exclusion_queries.append(Q(login_required=True))

        # If requested, exclude pages that are excluded from navigation, and their descendants
        if instance.in_navigation:
            exclusion_queries.append(
                Q(in_navigation=False) | Q(in_navigation__isnull=True)
            )

        # On a public page, we should display the public version only if it is
        # published in the current language
        if not current_page_is_draft:
            exclusion_queries.append(
                Q(title_set__language=language, title_set__published=False)
                | Q(
                    title_set__language=language,
                    title_set__publisher_state=PUBLISHER_STATE_PENDING,
                )
            )

        # We use a regex query on the path to exclude all the descendants of the pages we
        # identified as pages to exclude from the sitemap
        if exclusion_queries:
            paths_to_exclude = pages.filter(
                reduce(lambda q1, q2: q1 | q2, exclusion_queries, Q())
            ).values_list("node__path", flat=True)

            if paths_to_exclude:
                exclusion_regex = r"^({:s}).*".format("|".join(paths_to_exclude))
                pages = pages.exclude(node__path__regex=exclusion_regex)

        # Annotate pages with "level" ie their depth relative to the root page
        pages = pages.annotate(level=F("node__depth") - depth + 1)

        context["annotated_pages"] = annotate_pages(pages)
        return context



"""
Large banner plugin forms
"""
from django import forms

from djangocms_text_ckeditor.widgets import TextEditorWidget

from .models import LargeBanner

CKEDITOR_CONFIGURATION_NAME = "CKEDITOR_SETTINGS"


class LargeBannerForm(forms.ModelForm):
    """
    Plugin form used to fill its content from frontend admin.
    """

    class Meta:
        """
        Form meta attributes
        """

        model = LargeBanner
        widgets = {
            "content": TextEditorWidget(configuration=CKEDITOR_CONFIGURATION_NAME)
        }
        fields = [
            "title",
            "background_image",
            "logo",
            "logo_alt_text",
            "template",
            "content",
        ]



"""
Glimpse CMS plugin
"""
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from richie.apps.core.defaults import PLUGINS_GROUP

from .models import Glimpse


@plugin_pool.register_plugin
class GlimpsePlugin(CMSPluginBase):
    """
    CMSPlugin to add a glimpse with a variant form factor
    """

    cache = True
    model = Glimpse
    module = PLUGINS_GROUP
    name = _("Glimpse")
    render_template = "richie/glimpse/glimpse.html"
    fieldsets = (
        (None, {"fields": ["title", "variant", "image", "content"]}),
        (
            _("Link settings"),
            {"classes": ("collapse",), "fields": (("link_url", "link_page"),)},
        ),
    )

    # pylint: disable=R0201
    def compute_variant(self, context, instance):
        """
        Get the right final "variant" value for template context.

        * If model instance has an empty "variant" attribute, use the "glimpse_variant"
          value from parent template context, default to None if not set;
        * If model instance has a not null "variant" attribute use it, no
          matter what the value from parent template context is;
        """
        return instance.variant or context.get("glimpse_variant")

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "placeholder": placeholder,
                "glimpse_variant": self.compute_variant(context, instance),
            }
        )
        return context



"""
API endpoints to access courses through ElasticSearch
"""
from django.conf import settings

from elasticsearch.exceptions import NotFoundError
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from .. import ES_CLIENT
from ..defaults import ES_PAGE_SIZE
from ..filter_definitions import FILTERS
from ..indexers import ES_INDICES
from ..utils.viewsets import AutocompleteMixin, ViewSetMetadata


class CoursesViewSet(AutocompleteMixin, ViewSet):
    """
    A simple viewset with GET endpoints to fetch courses
    See API Blueprint for details on consumer use
    """

    _meta = ViewSetMetadata(indexer=ES_INDICES.courses)

    # pylint: disable=no-self-use,unused-argument,too-many-locals
    def list(self, request, version):
        """
        Course search endpoint: build an ElasticSearch request from our query params so
        it searches its index and returns a list of matching courses
        """
        # Instantiate the form to allow validation/cleaning
        form_class = self._meta.indexer.form
        params_form = form_class(data=request.query_params)

        # Return a 400 with error information if the query params are not valid
        if not params_form.is_valid():
            return Response(status=400, data={"errors": params_form.errors})

        limit, offset, query, aggs = params_form.build_es_query()

        body = {
            "script_fields": params_form.get_script_fields(),
            "sort": params_form.get_sorting_script(),
        }

        # The querystring may request only the query or only the aggregations
        scope = params_form.cleaned_data["scope"]
        if form_class.OBJECTS in scope or not scope:
            body["query"] = query

        if form_class.FILTERS in scope or not scope:
            body["aggs"] = aggs

        # pylint: disable=unexpected-keyword-arg
        course_query_response = ES_CLIENT.search(
            _source=getattr(self._meta.indexer, "display_fields", "*"),
            index=self._meta.indexer.index_name,
            doc_type=self._meta.indexer.document_type,
            body=body,
            # Directly pass meta-params through as arguments to the ES client
            from_=offset,
            size=limit or getattr(settings, "RICHIE_ES_PAGE_SIZE", ES_PAGE_SIZE),
        )

        response_object = {
            "meta": {
                "count": len(course_query_response["hits"]["hits"]),
                "offset": offset,
                "total_count": course_query_response["hits"]["total"],
            }
        }
        if form_class.OBJECTS in scope or not scope:
            response_object["objects"] = [
                self._meta.indexer.format_es_object_for_api(es_course)
                for es_course in course_query_response["hits"]["hits"]
            ]

        if form_class.FILTERS in scope or not scope:
            filters = {
                name: faceted_definition
                for filter in FILTERS.values()
                for name, faceted_definition in filter.get_faceted_definitions(
                    course_query_response["aggregations"]["all_courses"],
                    data=params_form.cleaned_data,
                ).items()
            }
            response_object["filters"] = dict(
                sorted(filters.items(), key=lambda f: f[1]["position"])
            )

        # Will be formatting a response_object for consumption
        return Response(response_object)

    # pylint: disable=no-self-use,invalid-name,unused-argument
    def retrieve(self, request, pk, version):
        """
        Return a single course by ID
        """
        # Wrap the ES get in a try/catch to we control the exception we emit — it would
        # raise and end up in a 500 error otherwise
        try:
            query_response = ES_CLIENT.get(
                index=self._meta.indexer.index_name,
                doc_type=self._meta.indexer.document_type,
                id=pk,
            )
        except NotFoundError:
            return Response(status=404)

        # Format a clean course object as a response
        return Response(self._meta.indexer.format_es_object_for_api(query_response))



# Generated by Django 2.1.9 on 2019-06-18 06:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("cms", "0022_auto_20180620_1551")]

    operations = [
        migrations.CreateModel(
            name="HTMLSitemapPage",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="html_sitemap_htmlsitemappage",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "max_depth",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Limit the level of nesting that your sitemap will contain below this page. An empty field or 0 equals to no limit.",
                        null=True,
                        verbose_name="max depth",
                    ),
                ),
                (
                    "in_navigation",
                    models.BooleanField(
                        default=False,
                        help_text="Tick to exclude from sitemap the pages that are excluded from navigation.",
                        verbose_name="in navigation",
                    ),
                ),
                (
                    "include_root_page",
                    models.BooleanField(
                        default=True,
                        help_text="Tick to include the root page and its descendants. Untick to include only its descendants.",
                        verbose_name="include root page",
                    ),
                ),
                (
                    "root_page",
                    models.ForeignKey(
                        blank=True,
                        help_text='This page will be at the root of your sitemap (or its children if the "include root page" flag is unticked).',
                        limit_choices_to={"publisher_is_draft": True},
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="sitemaps",
                        to="cms.Page",
                        verbose_name="root page",
                    ),
                ),
            ],
            options={
                "verbose_name": "HTML Sitemap",
                "verbose_name_plural": "HTML Sitemaps",
            },
            bases=("cms.cmsplugin",),
        )
    ]



# Generated by Django 2.1.7 on 2019-02-21 01:18

from django.db import migrations, models

from ..defaults import LARGEBANNER_TEMPLATES


class Migration(migrations.Migration):

    dependencies = [("large_banner", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="largebanner",
            name="content",
            field=models.TextField(blank=True, default="", verbose_name="Content"),
        ),
        migrations.AddField(
            model_name="largebanner",
            name="template",
            field=models.CharField(
                choices=LARGEBANNER_TEMPLATES,
                default=LARGEBANNER_TEMPLATES[0][0],
                help_text="Choose template to render plugin.",
                max_length=150,
                verbose_name="Template",
            ),
        ),
    ]
